/* csv extension for PHP */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "php.h"
#include "ext/standard/info.h"
#include "ext/standard/php_string.h"
#include "php_csv.h"
#include "csv_arginfo.h"

#include <stdbool.h>
#include "zend_smart_str.h"
#include "Zend/zend_interfaces.h"

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(csv)
{
	php_info_print_table_start();
		php_info_print_table_header(2, "CSV support", "enabled");
		php_info_print_table_header(2, "Version", PHP_CSV_VERSION);
		php_info_print_table_header(2, "Author", "George Peter Banyard");
		php_info_print_table_header(2, "Bug reports", PHP_CSV_BUGREPORT);
		php_info_print_table_header(2, "Sponsor me", "https://github.com/sponsors/Girgias");
	php_info_print_table_end();
}
/* }}} */

#define EOL_SEQUENCE_AND_DELIMITER_AND_ENCLOSURE_CHECKS(delimiter_arg_num, enclosure_arg_num, eol_arg_num) {	\
	if (eol_sequence) {																							\
		/* Make sure that there is at least one character in string */											\
		if (ZSTR_LEN(eol_sequence) == 0) {																		\
			zend_argument_value_error(eol_arg_num, "EOL sequence cannot be empty");								\
			RETURN_THROWS();																					\
		}																										\
	} else {																									\
		eol_sequence = zend_string_init("\r\n", strlen("\r\n"), 0);												\
	}																											\
																												\
	if (delimiter) {																							\
		/* Make sure that there is at least one character in string */											\
		if (ZSTR_LEN(delimiter) == 0) {																			\
			zend_argument_value_error(delimiter_arg_num, "Delimiter cannot be empty");							\
			zend_string_release(eol_sequence);																	\
			RETURN_THROWS();																					\
		}																										\
		if (zend_string_equals(delimiter, eol_sequence)) {                                                      \
			zend_value_error("Delimiter cannot be identical to the EOL sequence");	                            \
			zend_string_release(eol_sequence);																	\
			RETURN_THROWS();																					\
		}																										\
	} else {																									\
		delimiter = ZSTR_CHAR(',');                                                                             \
	}																											\
																												\
	if (enclosure) {																							\
		if (ZSTR_LEN(enclosure) == 0) {																			\
			zend_argument_value_error(enclosure_arg_num, "Enclosure cannot be empty");							\
			zend_string_release(delimiter);																		\
			zend_string_release(eol_sequence);																	\
			RETURN_THROWS();																					\
		}																										\
		if (zend_string_equals(enclosure, eol_sequence)) {                                                      \
			zend_value_error("Enclosure cannot be identical to the EOL sequence");	                            \
			zend_string_release(eol_sequence);																	\
			zend_string_release(delimiter);																		\
			RETURN_THROWS();																					\
		}																										\
	} else {																									\
		enclosure = ZSTR_CHAR('"');                                                                             \
	}																											\
																												\
	/* Ensure delimiter and enclosure are different */															\
	if (zend_string_equals(delimiter, enclosure)) {                                                             \
		zend_value_error("Delimiter and enclosure cannot be identical");                                        \
		zend_string_release(eol_sequence);																		\
		zend_string_release(delimiter);																			\
		zend_string_release(enclosure);																			\
		RETURN_THROWS();																						\
	}																											\
}

/**
 * Follows RFC4180 https://tools.ietf.org/html/rfc4180
 * The terminology can be slightly confusing here as what PHP considers the 'enclosure' is what is used
 * to escape a field in the RFC.
 *
 * This only formats ONE row of a CSV file.
 */
static zend_string* hashtable_to_rfc4180_string(
		HashTable *fields, const zend_string *delimiter, const zend_string *enclosure, const zend_string *eol_sequence)
{
	uint32_t nb_fields;
	uint32_t fields_iterated = 0;
	zval *tmp_field_zval;
	smart_str row = {0};

	nb_fields = zend_hash_num_elements(fields);
	ZEND_HASH_FOREACH_VAL(fields, tmp_field_zval) {
		zend_string *tmp_field_str;
		zend_string *field_str = zval_get_tmp_string(tmp_field_zval, &tmp_field_str);
		bool escape_field = false;
		bool enclosure_within_filed = false;

		/*
		 * A field must be escaped (enclosed) if it contains the delimiter OR a Carriage Return (\r)
		 * OR a Line Feed (\n) OR the field escape sequence (i.e. enclosure parameter) OR the custom EOL sequence
		 */
		if (
			memchr(ZSTR_VAL(field_str), '\n', ZSTR_LEN(field_str))
			|| memchr(ZSTR_VAL(field_str), '\r', ZSTR_LEN(field_str))
			|| php_memnstr(ZSTR_VAL(field_str), ZSTR_VAL(delimiter), ZSTR_LEN(delimiter),
				ZSTR_VAL(field_str) + ZSTR_LEN(field_str))
			|| php_memnstr(ZSTR_VAL(field_str), ZSTR_VAL(eol_sequence), ZSTR_LEN(eol_sequence),
				ZSTR_VAL(field_str) + ZSTR_LEN(field_str))
		) {
			escape_field = true;
		}
		/* If the field escape sequence (i.e. enclosure parameter) is within the field it needs to be
		 * duplicated within the field. */
		if (php_memnstr(ZSTR_VAL(field_str), ZSTR_VAL(enclosure), ZSTR_LEN(enclosure),
			ZSTR_VAL(field_str) + ZSTR_LEN(field_str))) {
			escape_field = true;
			enclosure_within_filed = true;
		}

		if (escape_field) {
			smart_str_append(&row, enclosure);

			if (enclosure_within_filed) {
				/* Create replace string (twice the field escape sequence) */
				smart_str escaped_enclosure = {0};
				smart_str_append(&escaped_enclosure, enclosure);
				smart_str_append(&escaped_enclosure, enclosure);
				smart_str_0(&escaped_enclosure);

				zend_string *replace = php_str_to_str(ZSTR_VAL(field_str), ZSTR_LEN(field_str), ZSTR_VAL(enclosure),
						ZSTR_LEN(enclosure), ZSTR_VAL(escaped_enclosure.s), ZSTR_LEN(escaped_enclosure.s));

				smart_str_append(&row, replace);

				smart_str_free(&escaped_enclosure);
				zend_string_release(replace);
			} else {
				smart_str_append(&row, field_str);
			}

			smart_str_append(&row, enclosure);
		} else {
			smart_str_append(&row, field_str);
		}

		/* Only add the delimiter in between fields on the same row. */
		if (++fields_iterated != nb_fields) {
			smart_str_append(&row, delimiter);
		}

		/* Clear temporary variable */
		zend_tmp_string_release(tmp_field_str);
	} ZEND_HASH_FOREACH_END();
	/* Add the EOL sequence to indicate the end of the row. */
	smart_str_append(&row, eol_sequence);

	smart_str_0(&row);
	return row.s;
}

static zend_string* hashtable_collection_to_rfc4180_buffer(
		HashTable *collection, const zend_string *delimiter,
		const zend_string *enclosure, const zend_string *eol_sequence)
{
	size_t collection_index = 0;
	zval *fields = NULL;
	uint32_t nb_fields = 0;
	smart_str buffer = {0};

	ZEND_HASH_FOREACH_VAL(collection, fields) {
		uint32_t new_nb_fields = 0;

		/* Check that fields is an array */
		if (Z_TYPE_P(fields) != IS_ARRAY) {
			zend_type_error("Element %zu of the collection must be an array", collection_index);
			smart_str_free(&buffer);
			return NULL;
		}

		/* used to check if the number of fields is equal in each iteration */
		new_nb_fields = zend_hash_num_elements(Z_ARRVAL_P(fields));
		/* collection_index == 0 means we are at the first iteration so don't check */
		if (nb_fields != new_nb_fields && collection_index != 0) {
			zend_value_error("Element %zu of the collection contains %"PRIu32
			" fields compared to %"PRIu32" fields on previous rows",
				collection_index, new_nb_fields, nb_fields);
			smart_str_free(&buffer);
			return NULL;
		}
		nb_fields = new_nb_fields;
		collection_index++;

		zend_string *result = hashtable_to_rfc4180_string(Z_ARRVAL_P(fields), delimiter, enclosure, eol_sequence);
		smart_str_append(&buffer, result);
		zend_string_release(result);
	} ZEND_HASH_FOREACH_END();

	smart_str_0(&buffer);
	return buffer.s;
}

/* Iterator handling */
static zend_string* iterator_collection_to_rfc4180_buffer(
		zend_object_iterator *iterator,
		const zend_string *delimiter,
		const zend_string *enclosure, const zend_string *eol_sequence)
{
	size_t collection_index = 0;
	zval *fields = NULL;
	uint32_t nb_fields = 0;
	smart_str buffer = {0};

	/* Get a zend_object_iterator* from the object, for how to use see
	 * https://heap.space/xref/php-src/ext/spl/spl_iterators.c?r=9e775db0#spl_iterator_apply
	 */

	iterator->index = 0;
	/* Rewind iterator */
	if (iterator->funcs->rewind) {
		iterator->funcs->rewind(iterator);
		if (EG(exception)) {
			smart_str_free(&buffer);
			return NULL;
		}
	}

	while (iterator->funcs->valid(iterator) == SUCCESS) {
		uint32_t new_nb_fields = 0;

		if (EG(exception)) {
			smart_str_free(&buffer);
			return NULL;
		}

		fields = iterator->funcs->get_current_data(iterator);
		/* Check that fields is an array */
		if (Z_TYPE_P(fields) != IS_ARRAY) {
			zend_type_error("Element %zu of the collection must be an array", collection_index);
			smart_str_free(&buffer);
			return NULL;
		}

		/* used to check if the number of fields is equal in each iteration */
		new_nb_fields = zend_hash_num_elements(Z_ARRVAL_P(fields));
		/* collection_index == 0 means we are at the first iteration so don't check */
		if (nb_fields != new_nb_fields && collection_index != 0) {
			zend_value_error("Element %zu of the collection contains %"PRIu32
			" fields compared to %"PRIu32" fields on previous rows",
				collection_index, new_nb_fields, nb_fields);
			smart_str_free(&buffer);
			return NULL;
		}
		nb_fields = new_nb_fields;
		collection_index++;

		zend_string *result = hashtable_to_rfc4180_string(Z_ARRVAL_P(fields), delimiter, enclosure, eol_sequence);
		smart_str_append(&buffer, result);
		zend_string_release(result);

		/* Move iterator forward */
		iterator->index++;
		iterator->funcs->move_forward(iterator);

		if (EG(exception)) {
			smart_str_free(&buffer);
			return NULL;
		}
	}

	smart_str_0(&buffer);
	return buffer.s;
}

static HashTable* rfc4180_string_to_hashtable(
		char **buffer, const size_t length, const zend_string *delimiter,
		const zend_string *enclosure, const zend_string *eol_sequence)
{
	size_t index;
	HashTable *return_value = zend_new_array(8);

	bool in_escaped_field = false;

	/* Dereference buffer */
	char *row = *buffer;

	smart_str field_value = {0};
	size_t field_value_length = 0;

	/* Main loop to "tokenize" the row */
	for (index = 0; index < length; ++index) {
		/* Check for field escape sequence (i.e. enclosure) */
		if (php_memnstr(row, ZSTR_VAL(enclosure), ZSTR_LEN(enclosure), row + ZSTR_LEN(enclosure))) {
			row += ZSTR_LEN(enclosure);

			if (!in_escaped_field) {
				in_escaped_field = true;
				continue;
			}

			/* In an escaped field */
			if (php_memnstr(row, ZSTR_VAL(enclosure), ZSTR_LEN(enclosure), row + ZSTR_LEN(enclosure))) {
				in_escaped_field = true;
				goto consume;
			}

			in_escaped_field = false;
			continue;
		}

		/* Check for delimiter if not in an escaped field */
		if (
			!in_escaped_field
			&& php_memnstr(
				row,
				ZSTR_VAL(delimiter),
				ZSTR_LEN(delimiter),
				row + ZSTR_LEN(delimiter)
			)
		) {
			row += ZSTR_LEN(delimiter);

			/* Add nul terminating byte */
			smart_str_0(&field_value);
			zval tmp;
			ZVAL_STR_COPY(&tmp, field_value.s);
			zend_hash_next_index_insert(return_value, &tmp);

			smart_str_free(&field_value);

			field_value_length = 0;
			continue;
		}

		/* Check for End Of Line sequence when not in an escaped field */
		if (!in_escaped_field && php_memnstr(row, ZSTR_VAL(eol_sequence), ZSTR_LEN(eol_sequence),
				row + ZSTR_LEN(eol_sequence))) {
			row += ZSTR_LEN(eol_sequence);
			goto eol;
		}

		consume:

		smart_str_appendc(&field_value, row[0]);
		field_value_length++;
		row++;
	}

	eol:
	/* Add nul terminating byte */
	smart_str_0(&field_value);
	zval tmp;
	ZVAL_STR_COPY(&tmp, field_value.s);
	zend_hash_next_index_insert(return_value, &tmp);

	smart_str_free(&field_value);
	/* Update outer buffer position */
	*buffer = row;

	return return_value;
}

static HashTable* rfc4180_buffer_to_hashtable_collection(
		zend_string *buffer, const zend_string *delimiter, const zend_string *enclosure,
		const zend_string *eol_sequence, bool is_lax)
{
	HashTable *return_value = zend_new_array(8);

	char *start_position = ZSTR_VAL(buffer);
	char *current_position = ZSTR_VAL(buffer);
	size_t length = ZSTR_LEN(buffer);
	size_t buffer_row_nb = 1;
	uint32_t nb_fields = 0;

	while (current_position - start_position < (ptrdiff_t) length) {
		uint32_t new_nb_fields = 0;
		HashTable *row = rfc4180_string_to_hashtable(&current_position, length, delimiter, enclosure, eol_sequence);

		/* used to check if the number of fields is equal in each iteration */
		new_nb_fields = zend_hash_num_elements(row);
		/* buffer_row_nb == 1 means we are at the first iteration so don't check */
		if (new_nb_fields != nb_fields && buffer_row_nb != 1 && !is_lax) {
			zend_value_error("Buffer row %zu contains %"PRIu32" fields compared to %"PRIu32" fields on previous rows",
				buffer_row_nb, new_nb_fields, nb_fields);
			zend_hash_destroy(row);
			zend_hash_destroy(return_value);
			FREE_HASHTABLE(row);
			FREE_HASHTABLE(return_value);
			return NULL;
		}
		nb_fields = new_nb_fields;
		buffer_row_nb++;

		zval tmp;
		ZVAL_ARR(&tmp, row);
		zend_hash_next_index_insert(return_value, &tmp);
	}

	return return_value;
}

PHP_METHOD(CSV, arrayToRow)
{
	zend_string *delimiter = NULL;
	zend_string *enclosure = NULL;
	zend_string *eol_sequence = NULL;
	HashTable *fields;

	if (zend_parse_parameters(ZEND_NUM_ARGS(), "h|SSS", &fields, &delimiter, &enclosure, &eol_sequence) == FAILURE) {
		RETURN_THROWS();
	}

	EOL_SEQUENCE_AND_DELIMITER_AND_ENCLOSURE_CHECKS(2, 3, 4);

	RETVAL_STR(hashtable_to_rfc4180_string(fields, delimiter, enclosure, eol_sequence));
	zend_string_release(eol_sequence);
	zend_string_release(delimiter);
	zend_string_release(enclosure);
}

PHP_METHOD(CSV, collectionToBuffer)
{
	zend_string *delimiter = NULL;
	zend_string *enclosure = NULL;
	zend_string *eol_sequence = NULL;
	zend_string *buffer = NULL;
	zval *collection;

	ZEND_PARSE_PARAMETERS_START(1, 4)
		Z_PARAM_ITERABLE(collection)
		Z_PARAM_OPTIONAL
		Z_PARAM_STR(delimiter)
		Z_PARAM_STR(enclosure)
		Z_PARAM_STR(eol_sequence)
	ZEND_PARSE_PARAMETERS_END();

	EOL_SEQUENCE_AND_DELIMITER_AND_ENCLOSURE_CHECKS(2, 3, 4);

	if (Z_TYPE_P(collection) == IS_ARRAY) {
		buffer = hashtable_collection_to_rfc4180_buffer(Z_ARRVAL_P(collection),
			delimiter, enclosure, eol_sequence);
	} else {
		zend_class_entry *ce;
		zend_object_iterator *iterator;

		ZEND_ASSERT(Z_TYPE_P(collection) == IS_OBJECT);

		ce = Z_OBJCE_P(collection);

		ZEND_ASSERT(ce->iterator_funcs_ptr != NULL);
		ZEND_ASSERT(ce->get_iterator != NULL);

		iterator = ce->get_iterator(ce, collection, 0);
		if (EG(exception)) {
			goto end;
		}
		ZEND_ASSERT(iterator != NULL);
		buffer = iterator_collection_to_rfc4180_buffer(iterator, delimiter, enclosure, eol_sequence);

		/* Release iterator */
		zend_iterator_dtor(iterator);
	}

	end:
	/* Release strings */
	zend_string_release(eol_sequence);
	zend_string_release(delimiter);
	zend_string_release(enclosure);
	/* If an Error/Exception has been thrown */
	if (EG(exception)) {
		RETURN_THROWS();
	}
	RETURN_STR(buffer);
}

PHP_METHOD(CSV, rowToArray)
{
	zend_string *delimiter = NULL;
	zend_string *enclosure = NULL;
	zend_string *eol_sequence = NULL;
	zend_string *row;

	if (zend_parse_parameters(ZEND_NUM_ARGS(), "S|SSS", &row, &delimiter, &enclosure, &eol_sequence) == FAILURE) {
		RETURN_THROWS();
	}

	EOL_SEQUENCE_AND_DELIMITER_AND_ENCLOSURE_CHECKS(2, 3, 4);

	char* row_c = ZSTR_VAL(row);
	RETVAL_ARR(rfc4180_string_to_hashtable(&row_c, ZSTR_LEN(row), delimiter, enclosure, eol_sequence));
	zend_string_release(eol_sequence);
	zend_string_release(delimiter);
	zend_string_release(enclosure);
}

static void buffer_to_collection_generic(INTERNAL_FUNCTION_PARAMETERS, bool is_lax)
{
	zend_string *delimiter = NULL;
	zend_string *enclosure = NULL;
	zend_string *eol_sequence = NULL;
	zend_string *buffer;

	if (zend_parse_parameters(ZEND_NUM_ARGS(), "S|SSS", &buffer, &delimiter, &enclosure, &eol_sequence) == FAILURE) {
		RETURN_THROWS();
	}

	EOL_SEQUENCE_AND_DELIMITER_AND_ENCLOSURE_CHECKS(2, 3, 4);

	HashTable *collection = rfc4180_buffer_to_hashtable_collection(buffer, delimiter, enclosure, eol_sequence, is_lax);

	/* Release strings */
	zend_string_release(eol_sequence);
	zend_string_release(delimiter);
	zend_string_release(enclosure);
	/* If an Error has been thrown */
	if (collection == NULL) {
		RETURN_THROWS();
	}
	RETURN_ARR(collection);
}

PHP_METHOD(CSV, bufferToCollection)
{
	buffer_to_collection_generic(INTERNAL_FUNCTION_PARAM_PASSTHRU, false);
}
PHP_METHOD(CSV, bufferToCollectionLax)
{
	buffer_to_collection_generic(INTERNAL_FUNCTION_PARAM_PASSTHRU, true);
}

/* CSV Static class definition */
static zend_class_entry *php_csv_ce = NULL;
static zend_object_handlers csv_object_handlers;

typedef struct _php_csv_object {
	zend_object std;
} php_csv_object;

static inline zend_object *csv_object_to_zend_object(php_csv_object *obj) {
	return ((zend_object*)(obj + 1)) - 1;
}

static inline php_csv_object *csv_object_from_zend_object(zend_object *zobj) {
	return ((php_csv_object*)(zobj + 1)) - 1;
}

static zend_object* csv_object_create(zend_class_entry* ce) {
	php_csv_object *obj = zend_object_alloc(sizeof(php_csv_object), ce);
	zend_object *zobj = csv_object_to_zend_object(obj);
	zend_object_std_init(zobj, ce);
	object_properties_init(zobj, ce);
	zobj->handlers = &csv_object_handlers;

	return zobj;
}

static zend_function *csv_object_get_constructor(zend_object *zobj) {
	zend_throw_error(NULL, "Cannot instantiate CSV class");
	return NULL;
}

/* {{{ ZEND_MINIT_FUNCTION */
ZEND_MINIT_FUNCTION(csv)
{
	php_csv_ce = register_class_CSV();
	php_csv_ce->create_object = csv_object_create;

	memcpy(&csv_object_handlers, &std_object_handlers, sizeof(zend_object_handlers));
	csv_object_handlers.offset = XtOffsetOf(php_csv_object, std);
	csv_object_handlers.get_constructor = csv_object_get_constructor;

	return SUCCESS;
}
/* }}} */

/* {{{ csv_module_entry
 */
zend_module_entry csv_module_entry = {
	STANDARD_MODULE_HEADER,
	"csv",					/* Extension name */
	NULL,			        /* zend_function_entry */
	ZEND_MINIT(csv),		/* PHP_MINIT - Module initialization */
	NULL,					/* PHP_MSHUTDOWN - Module shutdown */
	NULL,					/* PHP_RINIT - Request initialization */
	NULL,					/* PHP_RSHUTDOWN - Request shutdown */
	PHP_MINFO(csv),			/* PHP_MINFO - Module info */
	PHP_CSV_VERSION,		/* Version */
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_CSV
# ifdef ZTS
ZEND_TSRMLS_CACHE_DEFINE()
# endif
ZEND_GET_MODULE(csv)
#endif
